;;; lookup -- lookup word fragment in dictionary and replace with full word.

;; Author: Lawrence R. Dodd <dodd@roebling.poly.edu>
;; Maintainer: Lawrence R. Dodd <dodd@roebling.poly.edu>
;; Created: 1993 Sep 13
;; Version: 1.19
;; Keywords: extensions

;;; Copyright (C) 1993 Lawrence R. Dodd

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:

;;; Idea based on spell-dict.el by John W. Peterson but uses an improved
;;; algorithm.

;;; VERSION:
;;;  
;;; !Id: lookup.el,v 1.19 1993/09/18 14:01:08 dodd Exp !	 
;;; !Date: 1993/09/18 14:01:08 !
;;; !Revision: 1.19 ! 
;;; 
;;; Available in /roebling.poly.edu:/pub/

;;; LCD Archive Entry:
;;; lookup|Lawrence R. Dodd|dodd@roebling.poly.edu|
;;; lookup word fragment in dictionary and replace with full word.|
;;; !Date: 1993/09/18 14:01:08 !|!Revision: 1.19 !|~/misc/lookup.el.Z|

;;; MOTIVATION: 
;;; 
;;; This is a feature missing from some spell checkers and is inconvenient to
;;; use in other spell checkers.  It allows you to get a completion list from
;;; the system dictionary for a word fragment. 
;;; 
;;;   Example 1: The word `Saskatchewan' needs to be spelled.  The user may
;;;   type `Sas' and hit M-? (assuming you bind `lookup-word' to M-?) and a
;;;   completion list will be built using the shell command `look' which can
;;;   then be viewed with TAB.  Based on this completion list the user can
;;;   locate the word, hit RETURN, and the word `Saskatchewan' will be
;;;   inserted in place of `Sas.'
;;;   
;;;   Example 2: The user has typed `Sasaquane' and M-$ (ispell-word) gives no
;;;   "near-misses" in which case you back up to `Sas' and hit M-?
;;;   (lookup-word), find the correct word, hit return, and the word
;;;   `Sasaquane' will be replaced by the correct `Saskatchewan'.
;;;   
;;;   Example 3: Word fragments may also be used for the search.  The word
;;;   `pneumonia' needs to be spelled.  All the user can remember is the `mon'
;;;   in which case she/he can do `C-u M-x lookup-word' (or C-u M-? if
;;;   lookup-word is bound to M-?) on `mon' and get a list of all words
;;;   containing the interior word fragment `mon.'
;;; 
;;; If the user wishes, `C-x o' can be used for stepping into the
;;; *Completions* buffer in order to find the desired word. 
;;; 
;;; A more sophisticated method is to use search-completions.el by Radey
;;; Shouman <rshouman@chpc.utexas.edu> which is available at the
;;; elisp-archive.  That package allows further regular expression searching
;;; of the *Completions* buffer.  A very useful package.

;;; NOTE: 
;;; 
;;; The version of ispell.el that comes with GNU Ispell 4.0 does have an `L'
;;; (regular expression look up) option for misspelled words.  Since this
;;; ispell.el feature can only be run after finding a misspelled words you can
;;; not do a lookup on `cat' for `catechetical' since `cat' is a correctly
;;; spelled word.  Furthermore, ispell.el does not return the entire list
;;; returned by `look' (for some reason).  I personally use both lookup.el and
;;; the `L' option of GNU Ispell's ispell.el and find they compliment each
;;; other.

;;; INSTALLATION:
;;;
;;; To use this package, simply put it in a file called "lookup.el" in a lisp
;;; directory known to Emacs (see the variable `load-path'), byte-compile it,
;;; and put the lines
;;;
;;;   (require 'lookup)
;;;   (define-key esc-map "?" 'lookup-word)
;;;
;;; in your ~/.emacs file or in the file default.el in the ../lisp directory.
;;; If you only want to load this when needed, then instead insert the lines
;;; 
;;;   (autoload 'lookup-word "lookup"
;;;     "Lookup word fragment in dictionary and replace with full word." t)
;;;   (define-key esc-map "?" 'lookup-word)
;;;  
;;; You may wish to bind to something other than M-?.  The user-defined
;;; variables are
;;; 
;;;  lookup-dict-file
;;;  lookup-regexp-ok
;;;  
;;; to find out more about these variables, load this file, put your cursor at 
;;; the end of any of the variable names, and hit C-h v [RET]. 
;;; 
;;; Also of interest: search-completions.el from the elisp-archive.

;;; ALGORITHM: 
;;;  
;;; Uses the shell command `look' to find words in the system dictionary and
;;; then parses the results for basic mini-buffer completion.  Can do a more
;;; robust search using a `look' with -r (regular expression) switch (e.g.,
;;; the `look' that comes with GNU ispell).  This is controlled by the
;;; variable `lookup-regexp-ok'.
;;; 
;;; This helps in cases (for example) where `look ca' for some reason yields
;;; nothing but `look -r "^ca.*"' yields many many words.  [Can someone
;;; explain to me why `look ca' yields nothing while `look ad' yields many?
;;; It seems like a bug in `look' to me since in both cases the word exists
;;; (i.e., both `ca' and `ad') but one gives a match and the other does not.
;;; All versions of `look' seem to do this.  Strange, huh?]
;;; 

;;; Code:

(defvar lookup-dict-file "/usr/local/lib/dict/words"
  "*Spelling dictionary file as string.
Used to override default dictionary that `look' searches.")

(defvar lookup-regexp-ok t
  "*t means `look' command can handle `-r' regular expression switch.
The version of `look' that comes with GNU ISpell has this feature.")

(defun lookup-word (&optional interior-frag)

  "Look up words that start with letters from point to word beginning.
Minibuffer completion is used to help search for word which is then inserted
into your buffer, replacing the word you had partially typed.  With optional
argument INTERIOR-FRAG (C-u \\[lookup-word]) the word fragment is assumed to
be an interior word fragment and `lookup-regexp-ok' needs to be t.  Type
\\[other-window] followed by \\[isearch-forward-regexp] or \\[isearch-forward]
to search `*Completions*' buffer by hand and \\[other-window] to return to
mini-buffer.  Type \\[keyboard-quit] to abort."

  (interactive "P")

  (let ((bow (save-excursion (forward-word -1) (point)))
        (eow (save-excursion (forward-word -1) (forward-word 1) (point)))
        lookup-string lookup-regexp to-be-replaced replacement lookup-list)

    ;; Get the string to look up and the word to replace.
    ;; Look up just part of a word, but replace all of it.
    (setq lookup-string (buffer-substring bow (point))
          to-be-replaced (buffer-substring bow eow)

          ;; Now define the regular expression for which we search. 
          ;; Make it an interior word fragment if necessary.
          lookup-regexp (if interior-frag
                            (concat "^.*" lookup-string ".*")
                          ;; Use fast binary search.
                          (concat "^" lookup-string)))

    ;; Create list of words in system dictionary starting with lookup-string.
    (save-excursion
      (message "Building list...")
      (set-buffer (get-buffer-create " *lookup*"))
      (erase-buffer)

      (if lookup-regexp-ok
          (call-process "look" nil t nil "-dfr" lookup-regexp lookup-dict-file)
        (call-process "look" nil t nil "-df" lookup-string lookup-dict-file))
            
      ;; Build list for completing-reading by storing each line in the alist
      ;; lookup-list starting from the bottom. 

      ;; I trashed my own similar method for doing this and use instead the
      ;; method used in `ispell-do-look' of ispell.el.  That way if someone
      ;; ever merges these two it will be easier. -lrd 9/15/93.

      (goto-char (point-min))
      (while (not (= (point-min) (point-max)))
	(end-of-line)
        (setq lookup-list (cons (buffer-substring (point-min) (point))
                                lookup-list))
	(forward-line)
	(delete-region (point-min) (point)))

      ;; Make the list into an alist.
      (setq lookup-list (mapcar 'list (reverse lookup-list)))

      ;; Clean.
      (erase-buffer)
      (message "Building list...done"))

    ;; Now use basic mini-buffer completion.
    (setq replacement
          (let ((completion-ignore-case t))
            (if (and interior-frag lookup-regexp-ok)
                (completing-read (concat "Word [" lookup-regexp "]? ") lookup-list)
              (completing-read "Word? " lookup-list nil nil lookup-string))))

    ;; Use search-replace so case is preserved.
    (goto-char eow) ; does nothing if already at word end.
    (search-backward to-be-replaced bow)
    (replace-match replacement nil)))

;;; provide this package
(provide 'lookup)

;;; lookup.el ends here
