;;; rmailgen.el --- generates guesses for RMAIL output files and FCC files.

;; Author: Lawrence R. Dodd <dodd@roebling.poly.edu>
;; Maintainer: Lawrence R. Dodd <dodd@roebling.poly.edu>
;; Created: 13 Jul 1993
;; Version: 1.32
;; Date: 1994/04/19 20:34:44
;; Keywords: rmail extensions

;; Copyright (C) 1993, 1994 Free Software Foundation, Inc.
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 1, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:

;;; LCD Archive Entry:
;;; rmailgen|Lawrence R. Dodd|dodd@roebling.poly.edu|
;;; Generates guesses for RMAIL output files and FCC files.|
;;; 1994/04/19 20:34:44|1.32|~/misc/rmailgen.el.Z|

;;; This package generates guesses for `rmail-output-to-rmail-file' (`o') and
;;; if `rmail-generate-archive-file-name' is non-nil (the default) it will
;;; also generate archive file names (i.e., the FCC file names), when replying
;;; to mail messages based on the header of the message.  This package also
;;; allows user to set a default mail directory (`rmail-default-directory').
;;;
;;; It parses the account names in the "From:" and "To:" lines of the
;;; original mail message.  It uses an expression-driven form of
;;; `rmail-output-file-alist' (this requires GNU Emacs version 19.17 or
;;; later).
;;;
;;; Since C-o (`rmail-output') does not use `rmail-output-file-alist' no name
;;; is automatically generated in this case.  However, since `o' is now smart
;;; enough to distinguish between Un*x and BABYL format there is really no
;;; need ever to use C-o.

;;; Bugs:
;;;
;;; If you encounter a bug in this code, or wish to suggest an enhancement,
;;; type
;;;  
;;;              M-x rmail-generate-submit-report
;;;  
;;; to set up an outgoing mail buffer, with the proper address to the
;;; `rmailgen.el' maintainer automatically inserted in the `To:' field.  This
;;; command also inserts information that the maintainer can use to recreate
;;; your exact setup, making it easier to verify your bug.

;;; Available:
;;;
;;; via anonymous ftp to roebling.poly.edu [128.238.5.31]
;;;  
;;;        /anonymous@roebling.poly.edu:/pub/lisp/rmailgen.el.gz
;;;  
;;; and at the elisp archive
;;;  
;;; /archive.cis.ohio-state.edu:/pub/gnu/emacs/elisp-archive/misc/rmailgen.Z

;;; Installation:
;;;
;;; If not already installed, save this as rmailgen.el in a directory known to
;;; Emacs (check your `load-path' variable for a list of such files).  In your
;;; ~/.emacs file you will need to
;;;  
;;;             1. change rmail-default-directory       (optional)
;;;             2. define mail-archive-file-name        (optional)
;;;             3. set rmail-generate-archive-file-name (optional)
;;;             4. set rmail-generate-downcase          (optional)
;;;             5. set rmail-output-file-alist          (optional)
;;;             6. load rmailgen
;;; 
;;; Here is an example.  You can copy this into your ~/.emacs and remove the
;;; `;;; ' prefix.

;;; ;; 1. I want to define a mail directory that isn't `~/'
;;; (setq rmail-default-directory "~/Mail/") ; must end in slash
;;; 
;;; ;; 2. I want to define my own default archive file name.
;;; ;; Regular `send-mail' will use this.
;;; (add-hook 'mail-mode-hook
;;;           (function (lambda ()
;;;                       (setq mail-archive-file-name
;;;                             "~/Mail/.CarbonCopy"))))
;;; 
;;; ;; 3. By default, if mail-archive-file-name is non-nil then
;;; ;; archive file names will be generated automatically based on
;;; ;; the message to which a reply is being constructed.  If I
;;; ;; wanted to turn this off I would put in a statement like
;;; ;; (setq rmail-generate-archive-file-name nil) here.
;;;  
;;; ;; 4. By default, rmailgen.el downcases generated filenames
;;; ;; If I wanted uppercase I would put in a statement like
;;; ;; (setq rmail-generate-downcase nil) here.
;;;  
;;; ;; 5. Add my own personal output list for specific friends 
;;; ;; and special subjects.
;;; 
;;; ;; First define rmail-output-file-alist, just in case this
;;; ;; is not already defined.  That is, may be
;;; ;; ../lisp/rmailout.el has not been loaded yet.
;;; (if (not (boundp 'rmail-output-file-alist))
;;;     (defvar rmail-output-file-alist nil))
;;; 
;;; (setq rmail-output-file-alist
;;;       (append
;;;        (list
;;; 
;;;         ;; For my friends (some have strange account names).
;;;         '("^From:[ \t]*.*jones.*" . "jmjones")
;;;         '("^From:[ \t]*.*Joe[ \t]*Smith.*" . "joe")
;;; 
;;;         ;; Special subject lines.
;;;         '("^Subject:[ \t]*.*crypt.*" . "crypt++")
;;;         '("^Subject:[ \t]*.*rmailgen.*" . "rmailgen")
;;;         '("^Subject:[ \t]*.*GNU Emacs 19 RMAIL Poll.*" . "rmail")
;;; 
;;;         ;; Add more entries here...
;;;         )
;;; 
;;;        ;; In case rmail-output-file-alist has been defined
;;;        ;; already elsewhere.
;;;        rmail-output-file-alist))
;;; 
;;; ;; 6. Load package.
;;; (require 'rmailgen)

;;; Code:

;;; User-defined variables.

(defvar rmail-generate-archive-file-name t
  "*If non-nil, generate archive filename using message being replied to.
`mail-archive-file-name' must be non-nil as well.  If t then guess
archive filename using `rmail-output-to-rmail-file' algorithm.  If not t, but
still non-nil, will use only explicitly mapped entries appearing in
`rmail-output-file-alist'")

(defvar rmail-generate-downcase t
  "*If non-nil, generate lowercase filenames.  Otherwise uppercase.")

;;; There are no interactive functions.  There is no need to read any further.
;;; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 
;;; It would be nice if there were an `rmail-load-hook'.  That way the
;;; installation of rmailgen.el could be done when rmail is first loaded --
;;; just like dired-x.el is loaded in dired-mode.


;;;; SECTION A.
;;;; Default mail directory stuff.

;;; Override all defaults defined by (or to be defined) by ../lisp/rmail.el.
;;; The names of these variables will be changed in GNU Emacs v19.23 so
;;; they will have to be changed here as well.

(defvar rmail-default-directory "~/"
  "Default directory ending in slash used to define rmail in and out defaults.
If non-nil then `rmail-file-name', `rmail-last-rmail-file', and
`rmail-last-file' are all prepended with this variable.  If nil, nothing is
done to these variables.")

(if rmail-default-directory
    (setq rmail-file-name (expand-file-name
                           (concat
                            rmail-default-directory
                            (if (and (boundp 'rmail-file-name)
                                     rmail-file-name)
                                (file-name-nondirectory rmail-file-name)
                              "RMAIL")))
          rmail-last-rmail-file (expand-file-name
                                 (concat
                                  rmail-default-directory
                                  (if (and (boundp 'rmail-last-rmail-file)
                                           rmail-last-rmail-file)
                                      (file-name-nondirectory
                                       rmail-last-rmail-file)
                                    "XMAIL")))
          rmail-last-file (expand-file-name
                           (concat
                            rmail-default-directory
                            (if (and (boundp 'rmail-last-file)
                                     rmail-last-file)
                                (file-name-nondirectory rmail-last-file)
                              "XMAIL")))))

;;;; SECTION B.
;;;; Generate a guess for `rmail-output-to-rmail-file' ("o" in RMAIL).

;;; Algorithm:
;;;
;;; 1. What ../lisp/rmailout.el of GNU Emacs 19 does:
;;;  
;;;    Each element of the variable `rmail-output-file-alist' is of the form
;;;    (REGEXP . NAME-EXP).  `rmail-output-to-rmail-file' searches the current
;;;    RMAIL message for REGEXP.  If found it then evaluates the NAME-EXP
;;;    although it may just be a plain string.  If no match is found then the
;;;    name of the last output file is used.
;;;
;;; 2. What this file does:
;;;  
;;;    REGEXP is defined so it will locate the login-name of the sender of the
;;;    current message by looking in the "From" field and then the "To" field.
;;;    NAME-EXP is the function `rmail-generate-output-name' which extracts this
;;;    name.  The sender's login-name is returned as a lower-case string as
;;;    long as it does not match the user's own login-name, otherwise it
;;;    returns nil so `rmail-output-to-rmail-file' will continue to search the
;;;    alist.

;;; Define `rmail-output-file-alist' if not already.  That is, maybe
;;; ../lisp/rmailout.el has not been loaded yet and user has not defined any
;;; values.  The value and the doc-string will both be overwritten when
;;; necessary this just prevents an error message when we try to `append' to
;;; this variable.

(if (not (boundp 'rmail-output-file-alist))
    (defvar rmail-output-file-alist nil))

;;; Regular expressions for three types of mail addresses have been moved
;;; together into the function `rmail-generate-output-name'.  This saves time since
;;; only one regexp search is used where as before up to six were used.
;;; Someone suggested something similar (thanks) but I have since lost their
;;; address.

;;; Internal variable. 
(defvar rmail-generate-alist
  ;; Parse `From' line of mail messages.
  (list
   '("^From:[ \t\"]*" . (rmail-generate-output-name))
   ))

;;; Attach ourselves to alist variable so that rmail-output finds us.
(setq rmail-output-file-alist
      (append
       rmail-output-file-alist
       rmail-generate-alist))

;;; This is the NAME-EXP of `rmail-output-file-alist'.  In this case it is an
;;; expression that returns a folder-name as a string.

(defun rmail-generate-output-name (&optional second-pass)

  ;; Returns suggested output or carbon-copy folder-name as lower-case string
  ;; or nil if string the same as user login name or contains a `/'.  Will
  ;; check `To:' line if `From:' line doesn't reveal any guesses.

  (let ((case-fold-search t) ; ignore case
        folder-name)

    ;; Test for the three types of addresses.
    (if (or

         ;; 1. Joe Blow <foo@bar.gar.edu>
         ;;    Retrieves login name `foo' in first parenthesized expression.
         (looking-at "[^\n<]+<\\([^ \t\n\f@%()<>\"]+\\)[@%]\\([^ \t\n\f<>()]+\\)>")

         ;; 2. foo@bar.gar.edu (Joe Blow)
         ;;    Retrieves username in first parenthesized expression.
         (looking-at "\\([^ \t\n\f@%()<>\"]+\\)\\([@%]\\([^ \t\n\f<>()]+\\)\\)?")

         ;; 3. <joe>
         ;;    Retrieves login name in first parenthesized expression.
         (looking-at "<\\([^>\"]+\\)>"))

        (progn

          ;; Extract folder-name.
          (setq folder-name (buffer-substring (match-beginning 1)
                                              (match-end 1)))

          ;; Check folder-name.
          (cond ((string-match (user-login-name) folder-name)

                 ;; User login and folder-name match.  Message may be
                 ;; carbon-copy of outgoing one.  If `CC:' (or `BCC:') used in
                 ;; sending, user's name will appear in `From:' line of copy
                 ;; sent back to user.  Parse contents of `To:' line.
                 (if (not second-pass) ; Return nil if second pass.
                     (save-excursion
                       (goto-char (point-min))
                       (if (re-search-forward "^To:[ \t]*" nil t)
                           ;; Call this defun one more time with non-nil arg.
                           (rmail-generate-output-name t)))))
               
                ((string-match "/" folder-name)
                 ;; Contains a `/', return nil.  For strange usernames like
                 ;; `company/G=JOE/I=M/S=BLOW' easier to set explicit output.
                 nil)

                ;; Return lower-case form.
                (t
                 ;; Use generated name.
                 (rmail-generate-case folder-name)))))))


;;;; SECTION C.
;;;; Generate an FCC filename based on "From" and "To" fields when replying to
;;;; a message.

;;; Algorithm:
;;;
;;; Part of `mail-setup-hook' as called by `mail-setup' of ../lisp/sendmail.el.
;;;
;;; If replying to a message then `mail-reply-buffer' will be non-nil.  Go to
;;; that buffer and use `rmail-output-file-alist' to search the "From" and
;;; "To" fields for a suggestion for an archive file-name (see above).
;;;
;;; If suggestion found then search mail buffer, which has already been setup
;;; by `mail-setup', for the old FCC field which will contain the value of
;;; `mail-archive-file-name'.  Remove this and replace with the suggestion
;;; prepended with `rmail-default-directory'.

(defun mail-get-archive-file-name ()

  ;; Inserts guess for archive file name, based on contents variable
  ;; `rmail-output-file-alist', in mail buffer if `mail-archive-file-name' and
  ;; `rmail-generate-archive-file-name' are non-nil and if replying to a
  ;; message.  Part of `mail-setup-hook'.
  (if (and mail-archive-file-name
           rmail-generate-archive-file-name
           mail-reply-buffer)
      (let (answer
            ;; Generate guess for archive name only if
            ;; rmail-generate-archive-file-name is t.
            (tail
             (if (eq rmail-generate-archive-file-name t)
                 rmail-output-file-alist
               (delete rmail-generate-alist rmail-output-file-alist))))
        (save-excursion
          ;; Search the rmail buffer.
          (set-buffer mail-reply-buffer)
          ;; Suggest a file based on a pattern match.
          (while (and tail (not answer))
            (save-excursion
              (goto-char (point-min))
              (if (re-search-forward (car (car tail)) nil t)
                  (setq answer (eval (cdr (car tail)))))
              (setq tail (cdr tail)))))
        ;; If we found a suggestion, then insert it in the mail buffer.
        (if answer
            (save-excursion
              (goto-char (point-min))
              ;; Search and remove old FCC.  We know this is there because
              ;; `mail-archive-file-name' is non-nil so `mail-setup' has
              ;; inserted it already.
              (re-search-forward
               (concat "^FCC: " (regexp-quote mail-archive-file-name) "\n"))
              (delete-region (match-beginning 0) (match-end 0))
              ;; Insert new FCC.
              (insert "FCC: " rmail-default-directory answer "\n"))))))

;;; Put on hook.
(add-hook 'mail-setup-hook 'mail-get-archive-file-name)


;;;; SECTION D.
;;; Utilities 
(defun rmail-generate-case (obj)
  ;; Downcase OBJ if rmail-generate-downcase is non-nil, otherwise upcase.
  (if rmail-generate-downcase
      (downcase obj)
    (upcase obj)))


;;;; BUG REPORTS
;;; Uses Barry A. Warsaw's reporter.el

(defconst rmail-generate-version "1.32"
  "Revision number of rmailgen.el -- generates RMAIL output and FCC names.
Type \\[rmail-generate-submit-report] to send a bug report.  Available via
anonymous ftp in

   /roebling.poly.edu:/pub/lisp/rmailgen.gz
   /archive.cis.ohio-state.edu:/pub/gnu/emacs/elisp-archive/misc/rmailgen.Z")

(defun rmail-generate-submit-report ()
  "Submit via reporter.el a bug report on rmailgen."
  (interactive)

  ;; Since we have to be using v19 this should work.
  (require 'reporter)

  (reporter-submit-bug-report
   "dodd@roebling.poly.edu"                        ; address
   (concat "rmailgen.el " rmail-generate-version)  ; pkgname
   ;; varlist
   (list 'load-path ; useful
         'rmail-generate-downcase
         'rmail-generate-archive-file-name
         'rmail-output-file-alist        
         )
   nil nil                                         ; hooks
   (concat "Yo! Larry,")))                         ; salutation


;;;; Provide this package.
(provide 'rmailgen)

;;; rmailgen.el ends here
